/* Built-in files */

/* Generated with: stringify ../rc/joerc ../rc/ftyperc ../syntax/c.jsf ../colors/default.jcf ../syntax/comment_todo.jsf ../syntax/filename.jsf ../syntax/context.jsf */

#include "types.h"

const char *const builtins[]=
{
	"joerc",
		"-keepup	%k and %c status line escape sequences updated frequently\n"
		"-noxon		Attempt to turn off ^S/^Q processing\n"
		"-notite         Don't send terminal initialization and termination strings:\n"
		"-assume_color	Assume terminal has ANSI color support even if termcap/terminfo entry\n"
		"-assume_256color\n"
		"-guess_non_utf8	Allow guess of non-UTF-8 file encoding in a UTF-8 locale.\n"
		"-guess_utf16	Allow guess of UTF-16 encoding\n"
		"-guess_crlf     Automatically detect MS-DOS files and set -crlf appropriately\n"
		"-guess_indent	Guess indent character (tab or space).\n"
		"-menu_above	Position menu/list above prompt when enabled.  Otherwise position\n"
		"-transpose	Transpose rows with columns in all menus.\n"
		"-joe_state     Use ~/.joe_state file\n"
		"-brpaste	When JOE starts, send command to the terminal emulator that\n"
		"-pastehack	If keyboard input comes in as one block assume it's a mouse\n"
		"-restore	Restore previous cursor position when files are opened\n"
		"-search_prompting\n"
		"-lmsg \\i%k%T%W%I%X %n %m%y%R %M %x\n"
		"-rmsg  %S Row %4r Col %3c \n"
		"-smsg ** Line %r Col %c Offset %o(0x%O) %e %a(0x%A) Width %w ** \n"
		"-zmsg ** Line %r Col %c Offset %o(0x%O) ** \n"
		"-xmsg \\i Joe's Own Editor %v (%b) ** Type \\bCtrl-K Q\\b to exit or \\bCtrl-K H\\b for help **\\i\n"
		"-aborthint ^C\n"
		"-helphint ^K H\n"
		"-highlight\n"
		"-istep 4\n"
		":include ftyperc\n"
		":defmenu root\n"
		"mode,\"overwrite\",rtn	T Overtype %Zoverwrite%\n"
		"mode,\"hex\",rtn	' Hex edit mode\n"
		"mode,\"autoindent\",rtn	I Autoindent %Zautoindent%\n"
		"mode,\"wordwrap\",rtn	W Word wrap %Zwordwrap%\n"
		"mode,\"tab\",rtn	D Tab width %Ztab%\n"
		"mode,\"lmargin\",rtn	L Left margin %Zlmargin%\n"
		"mode,\"rmargin\",rtn	R Right margin %Zrmargin%\n"
		"mode,\"square\",rtn	X Column mode %Zsquare%\n"
		"mode,\"indentc\",rtn	% % Indent char %Zindentc%\n"
		"mode,\"istep\",rtn	% % Indent step %Zistep%\n"
		"menu,\"indent\",rtn	= Indent select\n"
		"mode,\"highlight\",rtn	H Highlighting %Zhighlight%\n"
		"mode,\"crlf\",rtn	Z CR-LF/MS-DOS %Zcrlf%\n"
		"mode,\"linums\",rtn	N Line numbers %Zlinums%\n"
		"mode,\"hiline\",rtn	U Highlight line %Zhiline%\n"
		"mode,\"beep\",rtn	B Beep %Zbeep%\n"
		"mode,\"rdonly\",rtn	O Read only %Zrdonly%\n"
		"mode,\"syntax\",rtn	Y Syntax\n"
		"mode,\"colors\",rtn	S Color scheme\n"
		"mode,\"encoding\",rtn	E Encoding\n"
		"mode,\"asis\",rtn	% % Meta chars as-is\n"
		"mode,\"language\",rtn	V Language\n"
		"mode,\"picture\",rtn	P picture %Zpicture%\n"
		"mode,\"type\",rtn		F File type [%Ztype%]\n"
		"mode,\"title\",rtn	C Context %Ztitle%\n"
		"menu,\"more-options\",rtn	  % % More options...\n"
		":defmenu more-options menu,\"root\",rtn\n"
		"menu,\"^G\",rtn	% % ^G options\n"
		"menu,\"search\",rtn	% % search options\n"
		"menu,\"paragraph\",rtn	% % paragraph options\n"
		"menu,\"file\",rtn	% % file options\n"
		"menu,\"menu\",rtn	% % menu options\n"
		"menu,\"global\",rtn	% % global options\n"
		"menu,\"cursor\",rtn	% % cursor options\n"
		"menu,\"marking\",rtn	% % marking options\n"
		"menu,\"tab\",rtn	% % tab/indent options\n"
		":defmenu indent menu,\"root\",rtn\n"
		"mode,\"istep\",rtn,\"1\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 1, Indent character = 32\",rtn	1 Space\n"
		"mode,\"istep\",rtn,\"2\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 2, Indent character = 32\",rtn	2 Spaces\n"
		"mode,\"istep\",rtn,\"3\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 3, Indent character = 32\",rtn	3 Spaces\n"
		"mode,\"istep\",rtn,\"4\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 4, Indent character = 32\",rtn	4 Spaces\n"
		"mode,\"istep\",rtn,\"5\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 5, Indent character = 32\",rtn	5 Spaces\n"
		"mode,\"istep\",rtn,\"8\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 8, Indent character = 32\",rtn	8 Spaces\n"
		"mode,\"istep\",rtn,\"10\",rtn,mode,\"indentc\",rtn,\"32\",rtn,msg,\"Indent step = 10, Indent character = 32\",rtn	0 Ten\n"
		"mode,\"istep\",rtn,\"1\",rtn,mode,\"indentc\",rtn,\"9\",rtn,msg,\"Indent step = 1, Indent character = 9\",rtn	T Tab\n"
		":defmenu menu menu,\"more-options\",rtn\n"
		"mode,\"menu_explorer\",rtn	% % Menu explorer %Zmenu_explorer%\n"
		"mode,\"menu_above\",rtn	% % Menu position %Zmenu_above%\n"
		"mode,\"menu_jump\",rtn	% % Jump into menu %Zmenu_jump%\n"
		"mode,\"transpose\",rtn	% % Transpose menus %Ztranspose%\n"
		":defmenu ^G menu,\"more-options\",rtn\n"
		"mode,\"highlighter_context\",rtn	% % ^G uses highlighter context %Zhighlighter_context%\n"
		"mode,\"single_quoted\",rtn	% % ^G ignores '...' %Zsingle_quoted%\n"
		"mode,\"no_double_quoted\",rtn	% % ^G no ignore \"...\" %Zno_double_quoted%\n"
		"mode,\"c_comment\",rtn	% % ^G ignores /*...*/ %Zc_comment%\n"
		"mode,\"cpp_comment\",rtn	% % ^G ignores //... %Zcpp_comment%\n"
		"mode,\"pound_comment\",rtn	% % ^G ignores #... %Zpound_comment%\n"
		"mode,\"vhdl_comment\",rtn	% % ^G ignores --... %Zvhdl_comment%\n"
		"mode,\"semi_comment\",rtn	% % ^G ignores ;... %Zsemi_comment%\n"
		"mode,\"tex_comment\",rtn % % ^G ignores %%... %Ztex_comment%\n"
		"mode,\"text_delimiters\",rtn % % Text delimiters %Ztext_delimiters%\n"
		":defmenu search menu,\"more-options\",rtn\n"
		"mode,\"icase\",rtn	% % Case insensitivity %Zicase%\n"
		"mode,\"wrap\",rtn	% % Search wraps %Zwrap%\n"
		"mode,\"search_prompting\",rtn	% % Search prompting %Zsearch_prompting%\n"
		"mode,\"csmode\",rtn	% % Continued search %Zcsmode%\n"
		":defmenu paragraph menu,\"more-options\",rtn\n"
		"mode,\"french\",rtn	% % French spacing %Zfrench%\n"
		"mode,\"flowed\",rtn	% % Flowed text %Zflowed%\n"
		"mode,\"cpara\",rtn	% % Paragraph indent chars %Zcpara%\n"
		"mode,\"cnotpara\",rtn	% % Not-paragraph chars %Zcnotpara%\n"
		":defmenu file menu,\"more-options\",rtn\n"
		"mode,\"restore\",rtn	% % Restore cursor %Zrestore%\n"
		"mode,\"guess_crlf\",rtn	% % Auto detect CR-LF %Zguess_crlf%\n"
		"mode,\"guess_indent\",rtn	% % Guess indent %Zguess_indent%\n"
		"mode,\"guess_non_utf8\",rtn	% % Guess non-UTF-8 %Zguess_non_utf8%\n"
		"mode,\"guess_utf8\",rtn	% % Guess UTF-8 %Zguess_utf8%\n"
		"mode,\"guess_utf16\",rtn	% % Guess UTF-16 %Zguess_utf16%\n"
		"mode,\"force\",rtn	% % Force last NL %Zforce%\n"
		"mode,\"nobackup\",rtn	% % No backup %Znobackup%\n"
		":defmenu global menu,\"more-options\",rtn\n"
		"mode,\"nolocks\",rtn	% % Disable locks %Znolocks%\n"
		"mode,\"nobackups\",rtn	% % Disable backups %Znobackups%\n"
		"mode,\"nodeadjoe\",rtn	% % Disable DEADJOE %Znodeadjoe%\n"
		"mode,\"nomodcheck\",rtn	% % Disable mtime check %Znomodcheck%\n"
		"mode,\"nocurdir\",rtn	% % Disable current dir %Znocurdir%\n"
		"mode,\"exask\",rtn	% % Exit ask %Zexask%\n"
		"mode,\"nosta\",rtn	% % Disable status line %Znosta%\n"
		"mode,\"keepup\",rtn	% % Fast status line %Zkeepup%\n"
		"mode,\"break_hardlinks\",rtn	% % Break hard links %Zbreak_hardlinks%\n"
		"mode,\"break_links\",rtn	% % Break links %Zbreak_links%\n"
		"mode,\"joe_state\",rtn	% % Joe_state file %Zjoe_state%\n"
		"mode,\"undo_keep\",rtn	% % No. undo records %Zundo_keep%\n"
		"mode,\"backpath\",rtn	% % Path to backup files %Zbackpath%\n"
		":defmenu cursor menu,\"more-options\",rtn\n"
		"mode,\"pg\",rtn	% % No. PgUp/PgDn lines %Zpg%\n"
		"mode,\"mid\",rtn	C Center on scroll %Zmid%\n"
		"mode,\"left\",rtn	L Columns to scroll left %Zleft%\n"
		"mode,\"right\",rtn R Columns to scroll right %Zright%\n"
		"mode,\"floatmouse\",rtn	% % Click past end %Zfloatmouse%\n"
		"mode,\"rtbutton\",rtn	% % Right button %Zrtbutton%\n"
		":defmenu marking menu,\"more-options\",rtn\n"
		"mode,\"autoswap\",rtn	% % Autoswap mode %Zautoswap%\n"
		"mode,\"marking\",rtn	% % Marking %Zmarking%\n"
		"mode,\"lightoff\",rtn	% % Auto unmask %Zlightoff%\n"
		":defmenu tab menu,\"more-options\",rtn\n"
		"mode,\"smarthome\",rtn	% % Smart home key %Zsmarthome%\n"
		"mode,\"smartbacks\",rtn	% % Smart backspace %Zsmartbacks%\n"
		"mode,\"indentfirst\",rtn	% % To indent first %Zindentfirst%\n"
		"mode,\"purify\",rtn	% % Clean up indents %Zpurify%\n"
		"mode,\"spaces\",rtn	% % No tabs %Zspaces%\n"
		"{Basic\n"
		"\\|\\uREGION\\u        \\|\\uGO TO\\u           \\|\\uGO TO\\u           \\|\\uDELETE\\u    \\|\\uEXIT\\u       \\|\\uSEARCH\\u    \\| \n"
		"\\|\\b^Arrow\\b Select \\|\\b^Z\\b Prev. word   \\|\\b^U/^V\\b PgUp/PgDn \\|\\b^D\\b Char.  \\|\\b^KX\\b Save   \\|\\b^KF\\b Find  \\| \n"
		"\\|\\b^KB\\b Begin     \\|\\b^X\\b Next word    \\|\\uMISC\\u            \\|\\b^Y\\b Line   \\|\\b^C\\b  Abort  \\|\\b^L\\b  Next  \\| \n"
		"\\|\\b^KK\\b End       \\|\\b^KU\\b Top of file \\|\\b^KJ\\b Paragraph   \\|\\b^W\\b >Word  \\|\\b^KQ\\b All    \\|\\uHELP\\u      \\| \n"
		"\\|\\b^KC\\b Copy      \\|\\b^KV\\b End of file \\|\\b^KA\\b Center line \\|\\b^O\\b Word<  \\|\\uFILE\\u       \\|\\bEsc .\\b Next\\| \n"
		"\\|\\b^KM\\b Move      \\|\\b^A\\b Beg. of line \\|\\b^K Space\\b Status \\|\\b^J\\b >Line  \\|\\b^KE\\b Edit   \\|\\bEsc ,\\b Prev\\| \n"
		"\\|\\b^KW\\b File      \\|\\b^E\\b End of line  \\|\\uSPELL\\u           \\|\\b^[O\\b Line< \\|\\b^KR\\b Insert \\|\\b^KH\\b Off   \\| \n"
		"\\|\\b^KY\\b Delete    \\|\\b^KL\\b To line no. \\|\\bEsc N\\b Word      \\|\\b^_\\b Undo   \\|\\b^KD\\b Save   \\|\\b^T\\b  Menu  \\| \n"
		"\\|\\b^K/\\b Filter    \\|\\b^G\\b  Matching (  \\|\\bEsc L\\b File      \\|\\b^^\\b Redo   \\|\\b^K`\\b Revert \\|\\b\\b          \\| \n"
		"}\n"
		"{Advanced\n"
		"\\|\\uMACROS\\u         \\|\\uSCROLL\\u           \\|\\uWINDOW\\u               \\|\\uGO TO\\u      \\|\\uI-SEARCH\\u     \\| \n"
		"\\|\\b^K[ 0-9\\b Record \\|\\bEsc W\\b Up         \\|\\b^KO\\b Open  \\b^C\\b  Close  \\|\\bEsc B\\b / \\bK\\b  \\|\\b^R\\b Backwards \\| \n"
		"\\|\\b^K]\\b     Stop   \\|\\bEsc Z\\b Down       \\|\\b^KG\\b Grow  \\b^KT\\b Shrink \\|to \\b^KB\\b/\\b^KK \\b\\|\\b^S\\b Forwards  \\| \n"
		"\\|\\bEsc I\\b   Text   \\|\\bEsc <\\b Left       \\|\\b^KN\\b Next  \\b^KP\\b Prev.  \\|\\b^P\\b  Up     \\|\\uBOOKMARKS\\u    \\| \n"
		"\\|\\b^K?\\b     Pause  \\|\\bEsc >\\b Right      \\|\\b^KI\\b Show one / All   \\|\\b^N\\b  Down   \\|\\bEsc 0-9\\b Goto \\| \n"
		"\\|   for prompt  \\|\\uINSERT\\u           \\|\\uSHELL\\u                \\|\\b^B\\b  Left   \\|\\bEsc Esc\\b Set  \\| \n"
		"\\|\\b^K 0-9\\b  Play   \\|\\b^]\\b Line  \\b^Q\\b Ctrl \\|\\bF1-F4\\b Term emulator  \\|\\b^F\\b  Right  \\|\\b^K-\\b Auto prev\\| \n"
		"\\|\\b^K\\\\\\b     Repeat \\|\\b^@\\b Space \\b^\\\\\\b Meta \\|\\b^K'\\b In buffer        \\|           \\|\\b^K=\\b Auto next\\| \n"
		"\\|\\bEsc D\\b   Dump   \\|\\bEsc Y\\b Paste      \\|\\b^KZ\\b Suspend  \\bEsc !\\b Append \\|      \\|             \\| \n"
		"}\n"
		"{Programs\n"
		"\\|\\uCOMMANDS\\u            \\|\\uCOMPILING\\u                  \\|\\uQUICK\\u                       \\| \n"
		"\\|\\bEsc X\\b Command prompt\\|\\bEsc C\\b Save modified compile\\|\\b^G\\b  Find word under cursor  \\| \n"
		"\\|\\bEsc M\\b Calculator    \\|\\bEsc G\\b Grep                 \\|\\b^K;\\b Tags file search        \\| \n"
		"\\|\\bshowlog\\b Startup log \\|\\bEsc =\\b To next error        \\|\\bEsc Enter\\b Complete word     \\| \n"
		"\\|\\bupper\\b   Change case \\|\\bEsc -\\b To prev. error       \\|\\uBUFFER\\u                      \\| \n"
		"\\|\\blower\\b   Change case \\|\\bEsc Space\\b Jump to error    \\|\\bEsc U\\b Next                  \\| \n"
		"\\|\\bcd\\b      Change dir  \\|\\uINDENT\\u                     \\|\\bEsc V\\b Previous              \\| \n"
		"\\|\\bkeymap\\b  Change map  \\|\\b^K,\\b Less                   \\|\\bEsc X bufed\\b Switch          \\| \n"
		"\\|                    \\|\\b^K.\\b More                   \\|                            \\| \n"
		"}\n"
		"{Search\n"
		"Search sequences:                                                            \\| \n"
		"    \\b\\\\^  \\\\$\\b  matches beg./end of line       \\b\\\\.\\b     match any single char      \\| \n"
		"    \\b\\\\<  \\\\>\\b  matches beg./end of word       \\b\\\\!\\b     match char or expression   \\| \n"
		"    \\b\\\\(  \\\\)\\b  grouping                       \\b\\\\|\\b     match left or right         \\| \n"
		"    \\b\\\\[a-z]\\b  matches one of a set           \\b\\\\y \\\\Y\\b  short for (.*) / (!*)       \\| \n"
		"    \\b\\\\{1,3}\\b  match 1 - 3 occurrences        \\b\\\\?\\b     match 0 or 1 occurrence     \\| \n"
		"    \\b\\\\+\\b      match 1 or more occurrences    \\b\\\\*\\b     match 0 or more occurrences \\| \n"
		"                                                                              \\|\n"
		"\\b^C\\b    to abort                             \\bEsc .\\b  to list of escape sequences \\|\n"
		"}\n"
		"{Escape sequences\n"
		"Escape sequences: \\| \n"
		"   \\b\\\\x{10fff}\\b Unicode code point    \\b\\\\p{Ll}\\b  Unicode category \\| \n"
		"   \\b\\\\i\\b / \\b\\\\I\\b   Identifier start      \\b\\\\t\\b  tab          \\b\\\\e\\b  escape               \\| \n"
		"   \\b\\\\c\\b / \\b\\\\C\\b   Identifier continue   \\b\\\\n\\b  newline      \\b\\\\r\\b  carriage return      \\| \n"
		"   \\b\\\\d\\b / \\b\\\\D\\b   Digit / Not a digit   \\b\\\\b\\b  backspace  \\b\\\\xFF\\b  hex character        \\| \n"
		"   \\b\\\\w\\b / \\b\\\\W\\b   Word / Not a word     \\b\\\\a\\b  alert      \\b\\\\377\\b  octal character      \\| \n"
		"   \\b\\\\s\\b / \\b\\\\S\\b   Space / Not a space   \\b\\\\f\\b  formfeed     \\b\\\\\\\\\\b  backslash            \\| \n"
		" \\|\n"
		" \\|\n"
		"}\n"
		"{Replace\n"
		"Replace sequences:                                                           \\| \n"
		"    \\b\\\\&\\b      replaced with entire match     \\b\\\\1\\b - \\b9\\b replaced with Nth group   \\| \n"
		"    \\b\\\\u \\\\l\\b   convert next to upper/lower    \\b\\\\U \\\\L\\b  case convert until \\b\\\\E\\b     \\| \n"
		"                                                                             \\|\n"
		"                                                                             \\|\n"
		"\\|\n"
		"\\|\n"
		"\\|\n"
		"\\b^C\\b    to abort                             \\bEsc ,\\b to list of escape sequences \\|\n"
		"}\n"
		"{SearchOptions\n"
		"Search options:                                                              \\| \n"
		"     \\br\\b Replace      \\bk\\b Restrict search to highlighted block                   \\| \n"
		"     \\bi\\b Ignore case  \\bb\\b Search backwards instead of forwards                   \\| \n"
		"                    \\ba\\b Search across all loaded files                         \\| \n"
		"                    \\be\\b Search across all files in Grep or Compile error list  \\| \n"
		" \\bw\\b / \\bn\\b  Allow / prevent wrap to start of file                                \\| \n"
		" \\bx\\b / \\by\\b  Search text is standard format / JOE format regular expression       \\| \n"
		"   \\bnnn\\b  Perform exactly nnn replacements                                     \\| \n"
		"\\b^C\\b    to abort\n"
		"}\n"
		"{Math\n"
		" \\uMATH COMMANDS\\u                  \\uCONSTANTS\\u                                \\| \n"
		" \\bhex oct bin\\b display mode       \\b0xff\\b hex entry                           \\| \n"
		" \\beng dec\\b engineering / normal   \\b0o77\\b octal entry                         \\| \n"
		"     \\bins\\b type result into file  \\b0b11\\b binary entry                        \\| \n"
		"    \\beval\\b evaluate block         \\b3e-4\\b floating point entry                \\| \n"
		" 2+3\\b:\\bins multiple commands      \\uVARIABLES\\u                                \\| \n"
		" \\uOPERATORS\\u                      \\bans\\b  Previous result                     \\| \n"
		" \\b! ^  * / %  + -\\b                \\ba=10\\b Assign a variable                   \\| \n"
		" \\b< <= > >= == !=  &&  ||  ? :\\b   \\bEsc .\\b for list of functions              \\| \n"
		"}\n"
		"{Functions\n"
		" \\uFUNCTIONS\\u                      \\uCONSTANTS\\u                                \\| \n"
		" \\bsin cos tab asin acos atan\\b     \\be pi\\b                                     \\| \n"
		" \\bsinh cosh tanh asinh acosh atanh\\b                                        \\| \n"
		" \\bsqrt cbrt exp ln log\\b                                                    \\| \n"
		" \\bint floor ceil abs\\b                                                      \\| \n"
		" \\berf erfc j0 j1 y0 y1\\b                                                    \\| \n"
		" \\bjoe(..macro..)\\b - runs an editor macro                                   \\| \n"
		"                                                                             \\| \n"
		"                                \\bEsc .\\b for statistics                     \\| \n"
		"}\n"
		"{Statistics \n"
		" \\uMATH COMMANDS\\u                                                               \\| \n"
		" Statistics on region of numbers  Linear regression on region of x y pairs   \\| \n"
		" \\bsum\\b  Sum                         \\blr(x)\\b  Estimate y given x     y=m*x+b      \\| \n"
		" \\bcnt\\b  Count                       \\brlr(y)\\b Estimate x given y                  \\| \n"
		" \\bavg\\b  Mean                        \\bLr rLr\\b Logarithmic regression y=m*ln(x)+b  \\| \n"
		"   Standard deviation:            \\blR rlR\\b Exponential regression y=e^(m*x)+b  \\| \n"
		" \\bdev\\b  all data present            \\bLR rLR\\b Power regression       y=b*x^m      \\| \n"
		" \\bsamp\\b sample of data present      Side effects: \\bb\\b, \\bm\\b: regression coef.       \\| \n"
		"                                      \\br\\b: correlation coef., \\bcov\\b: covariance  \\| \n"
		"}\n"
		"{Names\n"
		" Hit TAB to complete names.  Use up / down arrow keys to traverse history.   \\| \n"
		" Special file names:                                                         \\| \n"
		"      \\b!\\bcommand                 Pipe in/out of a shell command                \\| \n"
		"      \\b>>\\bfilename               Append to a file                              \\| \n"
		"      \\b-\\b                        Read/Write to/from standard I/O               \\|\n"
		"      filename\\b,START,SIZE\\b      Read/Write a part of a file/device            \\| \n"
		"          Give START/SIZE in decimal (255), octal (0377) or hex (0xFF)       \\| \n"
		"      \\b\\\\,\\b                       Use backslash to quote special characters     \\|\n"
		"\\b^C\\b    to abort\n"
		"}\n"
		"{CharTable\n"
		"\\| Dec  \\u 0123 4567  8901 2345    0123 4567  8901 2345 \\u  Dec \\| \n"
		"\\|   0 | \\u@ABC\\u \\uDEFG\\u  \\uHIJK\\u \\uLMNO\\u    \\i\\u@ABC\\u\\i \\i\\uDEFG\\u\\i  \\i\\uHIJK\\u\\i \\i\\uLMNO\\u\\i | 128 \\| \n"
		"\\|  16 | \\uPQRS\\u \\uTUVW\\u  \\uXYZ[\\u \\u\\\\]^_\\u    \\i\\uPQRS\\u\\i \\i\\uTUVW\\u\\i  \\i\\uXYZ[\\u\\i \\i\\u\\\\]^_\\u\\i | 144 \\| \n"
		"\\|  32 |  !\"# $%&'  ()*+ ,-./    \xa0\xa1\xa2\xa3 \xa4\xa5\xa6\xa7  \xa8\xa9\xaa\xab \xacX\xae\xaf | 160 \\| \n"
		"\\|  48 | 0123 4567  89:; <=>?    \xb0\xb1\xb2\xb3 \xb4\xb5\xb6\xb7  \xb8\xb9\xba\xbb \xbc\xbd\xbe\xbf | 176 \\| \n"
		"\\|  64 | @ABC DEFG  HIJK LMNO    \xc0\xc1\xc2\xc3 \xc4\xc5\xc6\xc7  \xc8\xc9\xca\xcb \xcc\xcd\xce\xcf | 192 \\| \n"
		"\\|  80 | PQRS TUVW  XYZ[ \\\\]^_    \xd0\xd1\xd2\xd3 \xd4\xd5\xd6\xd7  \xd8\xd9\xda\xdb \xdc\xdd\xde\xdf | 208 \\| \n"
		"\\|  96 | `abc defg  hijk lmno    \xe0\xe1\xe2\xe3 \xe4\xe5\xe6\xe7  \xe8\xe9\xea\xeb \xec\xed\xee\xef | 224 \\| \n"
		"\\| 112 | pqrs tuvw  xyz{ |}~\x7f    \xf0\xf1\xf2\xf3 \xf4\xf5\xf6\xf7  \xf8\xf9\xfa\xfb \xfc\xfd\xfe\xff | 240 \\| \n"
		"}\n"
		"{Joe\n"
		" Send bug reports to: http://sourceforge.net/projects/joe-editor \\| \n"
		" \\| \n"
		"  default joerc file is here /usr/local/etc/joe/joerc \\| \n"
		"  default syntax and i18n files are here /usr/local/share/joe \\| \n"
		"  additional documentation can be found here /usr/local/share/doc/joe \\| \n"
		" \\| \n"
		" \\| \n"
		" \\| \n"
		" \\| \n"
		"}\n"
		":def recenter retype,arg,\"top+height/2-line\",rtn,upslide\n"
		":windows		Bindings common to all windows\n"
		"type		U+0 TO U+10FFFF\n"
		"abort		^C		Abort window\n"
		"querysave,query,killjoe	^K Q	Query to save files, then exit\n"
		"querysave,query,killjoe	^K ^Q\n"
		"querysave,query,killjoe	^K q\n"
		"arg		^K \\		Repeat next command\n"
		"explode		^K I		Show all windows or show only one window\n"
		"explode		^K ^I\n"
		"explode		^K i\n"
		"help		^K H		Help menu\n"
		"help		^K ^H\n"
		"help		^K h\n"
		"hnext		^[ .		Next help window\n"
		"hprev		^[ ,		Previous help window\n"
		"math		^[ m		Calculator\n"
		"math		^[ M		Calculator\n"
		"msg		^[ h		Display a message\n"
		"msg		^[ H		Display a message\n"
		"msg		^[ ^H		Display a message\n"
		"nextw		^K N		Goto next window\n"
		"nextw		^K ^N\n"
		"nextw		^K n\n"
		"pgdn		.kN		Screen down\n"
		"pgdn		^V\n"
		"pgdn		^[ [ 6 ~\n"
		"pgup		.kP		Screen up\n"
		"pgup		^U\n"
		"pgup		^[ [ 5 ~\n"
		"play		^K 0 TO 9	Execute macro\n"
		"prevw		^K P		Goto previous window\n"
		"prevw		^K ^P\n"
		"prevw		^K p\n"
		"query		^K ?		Macro query insert\n"
		"record		^K [		Record a macro\n"
		"rtn		^M		Return\n"
		"shell		^K Z		Shell escape\n"
		"shell		^K ^Z\n"
		"shell		^K z\n"
		"stop		^K ]		Stop recording\n"
		"defmdown	MDOWN		Move the cursor to the mouse location\n"
		"defmup		MUP\n"
		"defmdrag	MDRAG		Select a region of characters\n"
		"defm2down	M2DOWN		Select the word at the mouse location\n"
		"defm2up		M2UP\n"
		"defm2drag	M2DRAG		Select a region of words\n"
		"defm3down	M3DOWN		Select the line at the mouse location\n"
		"defm3up		M3UP\n"
		"defm3drag	M3DRAG		Select a region of lines\n"
		"defmiddleup	MIDDLEUP\n"
		"defmiddledown	MIDDLEDOWN	Insert text\n"
		"xtmouse		^[ [ M		Introduces an xterm mouse event\n"
		"extmouse	^[ [ <		Introduces an extended xterm mouse event\n"
		":main			Text editing window\n"
		":inherit windows\n"
		":def ispellfile filt,\"SPLTMP=\\\"$(/bin/mktemp -p '' joeXXXXXX)\\\";cat >$SPLTMP;ispell $SPLTMP </dev/tty >/dev/tty;cat $SPLTMP;/bin/rm $SPLTMP\",rtn,retype\n"
		":def ispellword psh,nextword,markk,prevword,markb,filt,\"SPLTMP=\\\"$(/bin/mktemp -p '' joeXXXXXX)\\\"cat >$SPLTMP;ispell $SPLTMP </dev/tty >/dev/tty;tr -d <$SPLTMP '\\\\012';/bin/rm $SPLTMP\",rtn,retype,nextword\n"
		":def aspellfile filt,\"SPLTMP=\\\"$(/bin/mktemp -p '' joeXXXXXX)\\\";cat >$SPLTMP;aspell --lang=\",language,\".\",charset,\" -x -c $SPLTMP </dev/tty >/dev/tty;cat $SPLTMP;/bin/rm $SPLTMP\",rtn,retype\n"
		":def aspellword psh,nextword,markk,prevword,markb,filt,\"SPLTMP=\\\"$(/bin/mktemp -p '' joeXXXXXX)\\\";cat >$SPLTMP;aspell --lang=\",language,\".\",charset,\" -x -c $SPLTMP </dev/tty >/dev/tty;tr -d <$SPLTMP '\\\\012';/bin/rm $SPLTMP\",rtn,retype,nextword,pop\n"
		"aspellfile	^[ l\n"
		"aspellword	^[ n\n"
		":def compile mwind!,mfit!,querysave,query,scratch,\"* Build Log *\",rtn,bof,markb,eof,\" \",markk,blkdel,build\n"
		":def grep_find mwind!,mfit!,scratch,\"* Grep Log *\",rtn,bof,markb,eof,\" \",markk,blkdel,grep\n"
		":def man scratch,\"* Man Page *\",rtn,bof,markb,eof,\" \",markk,blkdel,\" \",ltarw,run,\"man -P cat -S 2:3 \"\n"
		":def shell1 scratch_push,\"* Startup Log *\",rtn,scratch_push,\"* Shell 1 *\",rtn,vtbknd!,eof\n"
		":def shell2 scratch_push,\"* Startup Log *\",rtn,scratch_push,\"* Shell 2 *\",rtn,vtbknd!,eof\n"
		":def shell3 scratch_push,\"* Startup Log *\",rtn,scratch_push,\"* Shell 3 *\",rtn,vtbknd!,eof\n"
		":def shell4 scratch_push,\"* Startup Log *\",rtn,scratch_push,\"* Shell 4 *\",rtn,vtbknd!,eof\n"
		":def shell_clear psh,bof,markb,eof,markk,blkdel\n"
		":def shell_parse parserr\n"
		":def shell_gparse gparse\n"
		":def shell_release release\n"
		":def shell_math maths\n"
		":def shell_abort abort\n"
		":def shell_typemath txt,maths,\"ins\",rtn,rtn,txt,\"\\r\",rtn,txt,quote,\"d\",rtn\n"
		":def shell_rtn rtn\n"
		":def shell_edit edit\n"
		":def shell_dellin dellin\n"
		":def shell_cd cd\n"
		":def shell_pop popabort\n"
		":def shell_markb markb\n"
		":def shell_markk markk\n"
		"paste			^[ ] 5 2 ;		Base64 paste (obsolete)\n"
		"brpaste			^[ [ 2 0 0 ~		Bracketed paste\n"
		"brpaste_done		^[ [ 2 0 1 ~		Bracketed paste done\n"
		"insc			^[ [ 2 ~\n"
		"insc			^[ [ L			SCO\n"
		"delch			^[ [ 3 ~\n"
		"pgup			^[ [ I		SCO\n"
		"pgdn			^[ [ G		SCO\n"
		"home			^[ [ 1 ~		Putty, Linux, Cygwin\n"
		"home			^[ [ H			Xterm, Konsole, SCO\n"
		"home			^[ O H			gnome-terminal\n"
		"home			^[ [ 7 ~		RxVT\n"
		"eol			^[ [ 4 ~		Putty, Linux, Cygwin, ssh\n"
		"eol			^[ [ F			Xterm, Konsole, SCO\n"
		"eol			^[ O F			gnome-terminal\n"
		"eol			^[ [ 8 ~		RxVT\n"
		"rtarw,ltarw,begin_marking,rtarw,toggle_marking	^[ [ 1 ; 5 C    Mark right Xterm\n"
		"rtarw,ltarw,begin_marking,rtarw,toggle_marking	^[ [ 5 C        Mark right Gnome-terminal\n"
		"rtarw,ltarw,begin_marking,rtarw,toggle_marking	^[ O c		Mark right RxVT Ctrl-rtarw\n"
		"ltarw,rtarw,begin_marking,ltarw,toggle_marking	^[ [ 1 ; 5 D    Mark left\n"
		"ltarw,rtarw,begin_marking,ltarw,toggle_marking	^[ [ 5 D        Mark left\n"
		"ltarw,rtarw,begin_marking,ltarw,toggle_marking	^[ O d		Mark left RxVT Ctrl-ltarw\n"
		"uparw,dnarw,begin_marking,uparw,toggle_marking	^[ [ 1 ; 5 A    Mark up\n"
		"uparw,dnarw,begin_marking,uparw,toggle_marking	^[ [ 5 A        Mark up\n"
		"uparw,dnarw,begin_marking,uparw,toggle_marking	^[ O a		Mark up RxVT Ctrl-uparw\n"
		"dnarw,uparw,begin_marking,dnarw,toggle_marking	^[ [ 1 ; 5 B    Mark down\n"
		"dnarw,uparw,begin_marking,dnarw,toggle_marking	^[ [ 5 B        Mark down\n"
		"dnarw,uparw,begin_marking,dnarw,toggle_marking	^[ O b		Mark down RxVT Ctrl-dnarw\n"
		"blkdel,nmark   ^[ [ 3 ; 5 -	Gnome-terminal Ctrl-Del\n"
		"blkdel,nmark   ^[ [ 3 ; 5 ~	Xterm Ctrl-Del\n"
		"yank		^[ [ 2 ; 5 -	Gnome-terminal Ctrl-Ins\n"
		"yank		^[ [ 2 ; 5 ~    Xterm Ctrl-Ins\n"
		"delbol		^[ o		Delete to beginning of line\n"
		"delbol		^[ ^O		Delete to beginning of line\n"
		"dnslide		^[ z		Scroll down one line\n"
		"dnslide		^[ Z		Scroll down one line\n"
		"dnslide		^[ ^Z		Scroll down one line\n"
		"dnslide,dnslide,dnslide,dnslide		MWDOWN\n"
		"compile		^[ c		Compile\n"
		"compile		^[ C\n"
		"compile		^[ ^C		Compile\n"
		"grep_find	^[ g		Grep\n"
		"grep_find	^[ G		Grep\n"
		"grep_find	^[ ^G		Grep\n"
		"execmd		^[ x		Prompt for command to execute\n"
		"execmd		^[ X		Prompt for command to execute\n"
		"execmd		^[ ^X		Prompt for command to execute\n"
		"finish		^[ ^I		Complete word in document\n"
		"finish		^[ ^M		Complete word: used to be math\n"
		"mwind!,mfit,jump,bol		^[ SP\n"
		"isrch		^[ s		Forward incremental search\n"
		"isrch		^[ S		Forward incremental search\n"
		"isrch		^[ ^S		Forward incremental search\n"
		"isrch		^S\n"
		"notmod		^[ ~		Not modified\n"
		"mwind,mfit,prevw,nxterr		^[ =		To next error\n"
		"parserr		^[ e		Parse errors in current buffer\n"
		"parserr		^[ E		Parse errors in current buffer\n"
		"parserr		^[ ^E		Parse errors in current buffer\n"
		"mwind,mfit,prevw,prverr		^[ -		To previous error\n"
		"rsrch		^[ r		Backward incremental search\n"
		"rsrch		^[ R		Backward incremental search\n"
		"rsrch		^[ ^R		Backward incremental search\n"
		"retype,rsrch	^R\n"
		"run		^[ !		Run a program in a window\n"
		"tomarkb		^[ b		Go to beginning of marked block\n"
		"tomarkb		^[ ^B		Go to beginning of marked block\n"
		"tomarkk		^[ k		Go to end of marked block\n"
		"tomarkk		^[ ^K		Go to end of marked block\n"
		"tomarkk		^[ K		Go to end of marked block\n"
		"txt		^[ i		Prompt for text and insert it\n"
		"txt		^[ I		Prompt for text and insert it\n"
		"upslide		^[ w		Scroll up one line\n"
		"upslide		^[ ^W		Scroll up one line\n"
		"upslide		^[ W		Scroll up one line\n"
		"upslide,upslide,upslide,upslide		MWUP\n"
		"yank		^[ y		Undelete text\n"
		"yankpop		^[ ^Y		Undelete text\n"
		"yank		^[ Y		Undelete text\n"
		"insc		^@		Ctrl-space used to insert a space\n"
		"nbuf		^[ v		Next buffer\n"
		"nbuf		^[ V		Next buffer\n"
		"nbuf		^[ ^V		Next buffer\n"
		"pbuf		^[ u		Prev buffer\n"
		"pbuf		^[ U		Prev buffer\n"
		"pbuf		^[ ^U		Prev buffer\n"
		"quote8		^\\		Enter Meta chars\n"
		"quote		^[ '		Enter Ctrl chars\n"
		"quote		^Q\n"
		"backs		^?		Backspace\n"
		"backs		^H\n"
		"backw		^O		Backspace word\n"
		"bknd		^K '		Shell window\n"
		"blkcpy		^K C		Copy marked block\n"
		"blkcpy		^K ^C\n"
		"blkcpy		^K c\n"
		"blkdel		^K Y		Delete marked block\n"
		"blkdel		^K ^Y\n"
		"blkdel		^K y\n"
		"blkmove		^K M		Move marked block\n"
		"blkmove		^K ^M\n"
		"blkmove		^K m\n"
		"blksave		^K W		Save marked block\n"
		"blksave		^K ^W\n"
		"blksave		^K w\n"
		"bof		^K U		Goto beginning of file\n"
		"bof		^K ^U\n"
		"bof		^K u\n"
		"home		.kh\n"
		"home		^A\n"
		"center		^K A		Center line\n"
		"center		^K ^A\n"
		"center		^K a\n"
		"crawll		^K <		Pan left\n"
		"crawll		^[ <		Pan left\n"
		"crawlr		^K >		Pan right\n"
		"crawlr		^[ >		Pan right\n"
		"delch		.kD		Delete character\n"
		"delch		^D\n"
		"deleol		^J		Delete to end of line\n"
		"dellin		^Y		Delete entire line\n"
		"delw		^W		Delete word to right\n"
		"dnarw		.kd		Go down\n"
		"dnarw		^N\n"
		"dnarw		^[ O B\n"
		"dnarw		^[ [ B\n"
		"edit		^K E		Edit a file\n"
		"edit		^K ^E\n"
		"edit		^K e\n"
		"eof		^K V		Go to end of file\n"
		"eof		^K ^V\n"
		"eof		^K v\n"
		"eol		.kH		Go to end of line\n"
		"eol		.@7\n"
		"eol		^E\n"
		"exsave		^K X		Save and exit\n"
		"exsave		^K ^X\n"
		"exsave		^K x\n"
		"ffirst		^K F		Find first\n"
		"ffirst		^K ^F\n"
		"ffirst		^K f\n"
		"filt		^K /		Filter block\n"
		"fnext		^L		Find next\n"
		"fmtblk		^K J		Format paragraphs in block\n"
		"fmtblk		^K ^J\n"
		"fmtblk		^K j\n"
		"gomark		^[ 0 TO 9	Goto bookmark\n"
		"groww		^K G		Grow window\n"
		"groww		^K ^G\n"
		"groww		^K g\n"
		"insc		.kI		Insert a space\n"
		"insf		^K R		Insert a file\n"
		"insf		^K ^R\n"
		"insf		^K r\n"
		"lindent		^K ,		Indent to left\n"
		"line		^K L		Goto line no.\n"
		"line		^K ^L\n"
		"line		^K l\n"
		"ltarw		.kl		Go left\n"
		"ltarw		^B\n"
		"ltarw		^[ O D\n"
		"ltarw		^[ [ D\n"
		"macros		^[ d		Dump keyboard macros\n"
		"macros		^[ ^D\n"
		"markb		^K B		Set beginning of marked block\n"
		"markb		^K ^B\n"
		"markb		^K b\n"
		"markk		^K K		Set end of marked block\n"
		"markk		^K ^K\n"
		"markk		^K k\n"
		"menu,\"root\",rtn	^T\n"
		"nextpos		^K =		Goto next position in position history\n"
		"nextword	^X		Goto next word\n"
		"open		^]		Split line\n"
		"prevpos		^K -		Previous position in history\n"
		"prevword	^Z		Previous word\n"
		"reload		^K `		Revert file\n"
		"redo		^^		Redo changes\n"
		"rindent		^K .		Indent to right\n"
		"rtarw		.kr		Go right\n"
		"rtarw		^F\n"
		"rtarw		^[ O C\n"
		"rtarw		^[ [ C\n"
		"run		^K !		Run a shell command\n"
		"save		^K D		Save file\n"
		"save		^K S\n"
		"save		^K ^D\n"
		"save		^K ^S\n"
		"save		^K d\n"
		"save		^K s\n"
		"setmark		^[ ^[		Set bookmark\n"
		"shrinkw		^K T		Shrink window\n"
		"shrinkw		^K ^T\n"
		"shrinkw		^K t\n"
		"splitw		^K O		Split window\n"
		"splitw		^K ^O\n"
		"splitw		^K o\n"
		"stat		^K SP		Show status\n"
		"stat		^K ^@		Show status\n"
		"tag		^K ;		Tags file search\n"
		"tomatch		^G		Goto matching parenthasis\n"
		"undo		^_		Undo changes\n"
		"uparw		.ku		Go up\n"
		"uparw		^P\n"
		"uparw		^[ O A\n"
		"uparw		^[ [ A\n"
		"shell1		^[ [ 1 1 ~\n"
		"shell1		^[ O P\n"
		"shell1		^[ [ [ A\n"
		"shell1		.k1\n"
		"shell2		^[ [ 1 2 ~\n"
		"shell2		^[ O Q\n"
		"shell2		^[ [ [ B\n"
		"shell2		.k2\n"
		"shell3		^[ [ 1 3 ~\n"
		"shell3		^[ O R\n"
		"shell3		^[ [ [ C\n"
		"shell3		.k3\n"
		"shell4		^[ [ 1 4 ~\n"
		"shell4		^[ O S\n"
		"shell4		^[ [ [ D\n"
		"shell4		.k4\n"
		":prompt			Prompt windows\n"
		":inherit main\n"
		"if,\"byte>size\",then,complete,complete,else,delch,endif	^D\n"
		"complete	^I\n"
		"dnarw,eol	.kd		Go down\n"
		"dnarw,eol	^N\n"
		"dnarw,eol	^[ O B\n"
		"dnarw,eol	^[ [ B\n"
		"uparw,eol	.ku		Go up\n"
		"uparw,eol	^P\n"
		"uparw,eol	^[ O A\n"
		"uparw,eol	^[ [ A\n"
		":menu			Selection menus\n"
		":inherit windows\n"
		"pgupmenu			^[ [ I\n"
		"pgdnmenu			^[ [ G\n"
		"bolmenu			^[ [ 1 ~		Putty, Linux, Cygwin\n"
		"bolmenu			^[ [ H			Xterm, Konsole\n"
		"bolmenu			^[ O H			gnome-terminal\n"
		"bolmenu			^[ [ 7 ~		RxVT\n"
		"eolmenu			^[ [ 4 ~		Putty, Linux, Cygwin, ssh\n"
		"eolmenu			^[ [ F			Xterm, Konsole\n"
		"eolmenu			^[ O F			gnome-terminal\n"
		"eolmenu			^[ [ 8 ~		RxVT\n"
		"abort		^[ ^[	Why?\n"
		"backsmenu	^?\n"
		"backsmenu	^H\n"
		"bofmenu		^K U\n"
		"bofmenu		^K ^U\n"
		"bofmenu		^K u\n"
		"bolmenu		.kh\n"
		"bolmenu		^A\n"
		"dnarwmenu	.kd\n"
		"dnarwmenu	^N\n"
		"dnarwmenu	^[ [ B\n"
		"dnarwmenu	^[ O B\n"
		"dnarwmenu	MWDOWN\n"
		"eofmenu		^K V\n"
		"eofmenu		^K ^V\n"
		"eofmenu		^K v\n"
		"eolmenu		.kH\n"
		"eolmenu		^E\n"
		"ltarwmenu	.kl\n"
		"ltarwmenu	^B\n"
		"ltarwmenu	^[ [ D\n"
		"ltarwmenu	^[ O D\n"
		"pgdnmenu	.kN		Screen down\n"
		"pgdnmenu	^V\n"
		"pgdnmenu	^[ [ 6 ~\n"
		"pgupmenu	.kP		Screen up\n"
		"pgupmenu	^U\n"
		"pgupmenu	^[ [ 5 ~\n"
		"rtarwmenu	.kr\n"
		"rtarwmenu	^F\n"
		"rtarwmenu	^[ [ C\n"
		"rtarwmenu	^[ O C\n"
		"rtn		SP\n"
		"rtn		^I\n"
		"rtn		^K H\n"
		"rtn		^K h\n"
		"rtn		^K ^H\n"
		"tabmenu		^I\n"
		"uparwmenu	.ku\n"
		"uparwmenu	^P\n"
		"uparwmenu	^[ [ A\n"
		"uparwmenu	^[ O A\n"
		"uparwmenu	MWUP\n"
		"defm2down	M2DOWN		Hits return key\n"
		":query			Single-key query window\n"
		":inherit windows\n"
		":querya			Single-key query window for quoting\n"
		"type		U+0 TO U+10FFFF\n"
		":querysr		Search & replace query window\n"
		"type		U+0 TO U+10FFFF\n"
		":shell			Input to shell windows\n"
		":inherit main\n"
		"\"\x03\"		^C		Abort\n"
		"\"\x04\"		^D		Eof\n"
		"\"\\t\"		^I		Tab\n"
		"\"\x08\"		^H		Backspace\n"
		"\"\\r\"		^M		Return\n"
		"\"\x7f\"		^?		Backspace\n"
		":vtshell		Input to ANSI shell windows\n"
		":inherit main\n"
		"\"\x1b[A\"		\x1b [ A\n"
		"\"\x1b[A\"		.ku\n"
		"\"\x1b[B\"		\x1b [ B\n"
		"\"\x1b[B\"		.kd\n"
		"\"\x1b[C\"		\x1b [ C\n"
		"\"\x1b[C\"		.kr\n"
		"\"\x1b[D\"		\x1b [ D\n"
		"\"\x1b[D\"		.kl\n"
		"\"\x01\"		^A		BOL for bash\n"
		"\"\x03\"		^C		Abort\n"
		"\"\x04\"		^D		Eof\n"
		"\"\x05\"		^E		EOL for bash\n"
		"\"\\t\"		^I		Tab\n"
		"\"\x08\"		^H		Backspace\n"
		"\"\\r\"		^M		Return\n"
		"\"\x7f\"		^?		Backspace\n"
,	"ftyperc",
		"[text]\n"
		"*\n"
		"*.txt\n"
		"[not_text]\n"
		"*.*\n"
		"-wordwrap\n"
		"[diff]\n"
		"*\n"
		"+\\? [A-Za-z]\n"
		"+Index: [A-Za-z]\n"
		"+Only in\n"
		"+Nur in\n"
		"+--- \n"
		"+\\*\\*\\* \n"
		"+[1-9][0-9]*[cda]\n"
		"*.diff\n"
		"*.patch\n"
		"-syntax diff\n"
		"[troff]\n"
		"*.1\n"
		"*.1.in\n"
		"-wordwrap\n"
		"-syntax troff\n"
		"[tex]\n"
		"*.tex\n"
		"*.sty\n"
		"-syntax tex\n"
		"-tex_comment\n"
		"-no_double_quoted\n"
		"-wordwrap\n"
		"[email]\n"
		"*.article*\n"
		"*.followup\n"
		"*.letter\n"
		"*tmp/nn.*\n"
		"*tmp/Re*\n"
		"*tmp/snd.*\n"
		"*tmp/dmt*\n"
		"*tmp/pico.*\n"
		"*tmp/mutt-*\n"
		"-wordwrap\n"
		"-syntax mail\n"
		"[avr]\n"
		"*.avr\n"
		"*.inc\n"
		"-syntax avr\n"
		"-autoindent\n"
		"-wordwrap\n"
		"-tab 8\n"
		"[assembly]\n"
		"*.asm\n"
		"*.s\n"
		"*.S\n"
		"-wordwrap\n"
		"-syntax asm\n"
		"[mason]\n"
		"*.mas\n"
		"-autoindent\n"
		"-title\n"
		"-syntax mason\n"
		"-smarthome\n"
		"-smartbacks\n"
		"[sml]\n"
		"*.sml\n"
		"-autoindent\n"
		"-title\n"
		"-syntax sml\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-istep 2\n"
		"[ocaml]\n"
		"*.ml\n"
		"*.mli\n"
		"-autoindent\n"
		"-title\n"
		"-syntax ocaml\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-istep 2\n"
		"[ruby]\n"
		"*.rb\n"
		"*\n"
		"+#![ 	]*[^\\n]*ruby[0-9.]*\\>\n"
		"*.gemspec\n"
		"*Gemfile\n"
		"*Rakefile\n"
		"*Guardfile\n"
		"*.rabl\n"
		"-autoindent\n"
		"-title\n"
		"-pound_comment\n"
		"-syntax ruby\n"
		"-text_delimiters do=end\n"
		"[perl]\n"
		"*\n"
		"+#![ 	]*[^\\n]*perl[0-9.]*\\>\n"
		"*.pl\n"
		"*.pm\n"
		"*.t\n"
		"-autoindent\n"
		"-title\n"
		"-syntax perl\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-pound_comment\n"
		"-single_quoted\n"
		"[sql]\n"
		"*.sql\n"
		"-autoindent\n"
		"-title\n"
		"-syntax sql\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"-c_comment\n"
		"-vhdl_comment\n"
		"-cpp_comment\n"
		"-text_delimiters BEGIN|Begin|begin=END|End|end\n"
		"-cpara >#!;*/%\n"
		"[awk]\n"
		"*.awk\n"
		"-autoindent\n"
		"-title\n"
		"-syntax awk\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-pound_comment\n"
		"[yacc]\n"
		"*.y\n"
		"-autoindent\n"
		"-title\n"
		"-syntax c\n"
		"-highlighter_context\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
		"-cpara >#!;*/%\n"
		"[lex]\n"
		"*.l\n"
		"*.lex\n"
		"-autoindent\n"
		"-title\n"
		"-syntax c\n"
		"-highlighter_context\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
		"-cpara >#!;*/%\n"
		"[ada]\n"
		"*.adb\n"
		"*.ads\n"
		"-syntax ada\n"
		"-autoindent\n"
		"-title\n"
		"-istep 2\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-vhdl_comment\n"
		"-text_delimiters declare|Declare|DECLARE|exception|Exception|EXCEPTION|if|If|IF|loop|Loop|LOOP|case|Case|CASE|package|Package|PACKAGE|procedure|Procedure|PROCEDURE|record|Record|RECORD|function|Function|FUNCTION=end|End|END\n"
		"[cobol]\n"
		"*.cbl\n"
		"*.cob\n"
		"-syntax cobol\n"
		"[rexx]\n"
		"*.rex\n"
		"*.REX\n"
		"*\n"
		"+#![ 	]*[^\\n]*rexx[0-9.]*\\>\n"
		"-syntax rexx\n"
		"-autoindent\n"
		"-title\n"
		"[sed]\n"
		"*.sed\n"
		"-syntax sed\n"
		"[postscript]\n"
		"*.ps\n"
		"*.eps\n"
		"-syntax ps\n"
		"[c]\n"
		"*.c\n"
		"*.cpp\n"
		"*.cc\n"
		"*.c++\n"
		"*.h\n"
		"*.h.in\n"
		"*.hpp\n"
		"*.h++\n"
		"*.hh\n"
		"*.mm\n"
		"-autoindent\n"
		"-title\n"
		"-syntax c\n"
		"-highlighter_context\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
		"-cpara >#!;*/%\n"
		"[verilog]\n"
		"*.v\n"
		"*.vh\n"
		"-autoindent\n"
		"-syntax verilog\n"
		"-istep 2\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-text_delimiters `ifdef|`ifndef=`else=`endif:begin=end:case|casex|casez=endcase:function=endfunction:module=endmodule:task=endtask:attribute=endattribute:primitive=endprimitive:table=endtable\n"
		"-cpara >#!;*/%\n"
		"[vhdl]\n"
		"*.vhd\n"
		"-autoindent\n"
		"-syntax vhdl\n"
		"-istep 2\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-vhdl_comment\n"
		"-text_delimiters entity|ENTITY|if|IF|component|COMPONENT|loop|LOOP|configuration|CONFIGURATION|units|UNITS|record|RECORD|case|CASE|function|FUNCTION|package|PACKAGE|architecture|ARCHITECTURE|block|BLOCK|process|PROCESS|generate|GENERATE=end|END\n"
		"-cpara >#!;*-/%\n"
		"[xml]\n"
		"*.xml\n"
		"*.xsd\n"
		"*\n"
		"+<.xml\n"
		"*.jnlp\n"
		"*.resx\n"
		"*build.xml\n"
		"-autoindent\n"
		"-syntax xml\n"
		"[plist]\n"
		"*.plist\n"
		"-autoindent\n"
		"-syntax xml\n"
		"[html]\n"
		"*.htm\n"
		"*.html\n"
		"-wordwrap\n"
		"-autoindent\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-syntax html\n"
		"[css]\n"
		"*.css\n"
		"-wordwrap\n"
		"-autoindent\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-syntax css\n"
		"[java]\n"
		"*.java\n"
		"-autoindent\n"
		"-title\n"
		"-syntax java\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-purify\n"
		"-cpara >#!;*/%\n"
		"[javascript]\n"
		"*.js\n"
		"*\n"
		"+#![ 	]*[^\\n]*nodejs[0-9.]*\\>\n"
		"-autoindent\n"
		"-title\n"
		"-syntax js\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-cpara >#!;*/%\n"
		"[json]\n"
		"*.json\n"
		"-autoindent\n"
		"-syntax json\n"
		"-smarthome\n"
		"-smartbacks\n"
		"[typescript]\n"
		"*.ts\n"
		"-autoindent\n"
		"-title\n"
		"-syntax typescript\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-cpara >#!;*/%\n"
		"[php]\n"
		"*.php\n"
		"-autoindent\n"
		"-title\n"
		"-syntax php\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"[python]\n"
		"*\n"
		"+#![ 	]*[^\\n]*python[0-9.]*\\>\n"
		"*.py\n"
		"-autoindent\n"
		"-title\n"
		"-syntax python\n"
		"-highlighter_context\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"-pound_comment\n"
		"[csh]\n"
		"*\n"
		"+#![ 	]*[^\\n 	]*/csh\\>\n"
		"+#![ 	]*[^\\n 	]*/tcsh\\>\n"
		"*.csh\n"
		"*.login\n"
		"*.logout\n"
		"*.tcsh\n"
		"*.tcshrc\n"
		"-autoindent\n"
		"-title\n"
		"-syntax csh\n"
		"-pound_comment\n"
		"[sh]\n"
		"*\n"
		"+#![ 	]*[^\\n 	]*/sh\\>\n"
		"+#![ 	]*[^\\n 	]*/bash\\>\n"
		"*.sh\n"
		"*profile\n"
		"*.bash\n"
		"*.bashrc\n"
		"*.bash_login\n"
		"*.bash_logout\n"
		"-autoindent\n"
		"-title\n"
		"-syntax sh\n"
		"-highlighter_context\n"
		"-pound_comment\n"
		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
		"[lisp]\n"
		"*.lisp\n"
		"*.lsp\n"
		"*.el\n"
		"-autoindent\n"
		"-title\n"
		"-syntax lisp\n"
		"-semi_comment\n"
		"[ksh]\n"
		"*\n"
		"+#![ 	]*[^\\n 	]*/ksh\\>\n"
		"*.ksh\n"
		"-autoindent\n"
		"-title\n"
		"-syntax sh\n"
		"-highlighter_context\n"
		"-pound_comment\n"
		"-text_delimiters do=done:if=elif=else=fi:case=esac\n"
		"[pascal]\n"
		"*.p\n"
		"*.pas\n"
		"-autoindent\n"
		"-title\n"
		"-syntax pascal\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-text_delimiters begin|BEGIN|record|RECORD|case|CASE=end|END:repeat|REPEAT=until|UNTIL\n"
		"[fortran]\n"
		"*.f\n"
		"*.f90\n"
		"*.for\n"
		"*.FOR\n"
		"-autoindent\n"
		"-title\n"
		"-syntax fortran\n"
		"[skill]\n"
		"*.il\n"
		"-semi_comment\n"
		"-syntax skill\n"
		"-autoindent\n"
		"-title\n"
		"-istep 3\n"
		"-indentc 32\n"
		"-smartbacks\n"
		"-smarthome\n"
		"-indentfirst\n"
		"[lua]\n"
		"*.lua\n"
		"*\n"
		"+#![ 	]*[^\\n]*lua[0-9.]*\\>\n"
		"-autoindent\n"
		"-title\n"
		"-syntax lua\n"
		"[tcl]\n"
		"*.tcl\n"
		"-autoindent\n"
		"-title\n"
		"-syntax tcl\n"
		"[go]\n"
		"*.go\n"
		"-autoindent\n"
		"-title\n"
		"-syntax go\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"-c_comment\n"
		"-text_delimiters #if|#ifdef|#ifndef=#elif=#else=#endif\n"
		"-cpara >#!;*/%\n"
		"[c#]\n"
		"*.cs\n"
		"-autoindent\n"
		"-title\n"
		"-syntax csharp\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-cpara >#!;*/%\n"
		"-text_delimiters #region=#endregion:#if=#elif=#else=#endif\n"
		"[jsf]\n"
		"*.jsf\n"
		"*.jsf.in\n"
		"-autoindent\n"
		"-title\n"
		"-syntax jsf\n"
		"-smarthome\n"
		"[jcf]\n"
		"*.jcf\n"
		"-syntax jcf\n"
		"[m4]\n"
		"*.ac\n"
		"*.m4\n"
		"-autoindent\n"
		"-title\n"
		"-syntax m4\n"
		"[matlab]\n"
		"*.m\n"
		"-wordwrap\n"
		"-syntax matlab\n"
		"[conf]\n"
		"*.am\n"
		"*rc\n"
		"*Makefile\n"
		"*makefile.in\n"
		"*Makefile.in\n"
		"*makefile\n"
		"*AKEFILE\n"
		"-autoindent\n"
		"-syntax conf\n"
		"-pound_comment\n"
		"[bat]\n"
		"*.bat\n"
		"*.cmd\n"
		"-autoindent\n"
		"-syntax batch\n"
		"[joerc]\n"
		"*ftyperc\n"
		"*joerc\n"
		"*jmacsrc\n"
		"*jstarrc\n"
		"*rjoerc\n"
		"*jpicorc\n"
		"*joerc.in\n"
		"*jmacsrc.in\n"
		"*jstarrc.in\n"
		"*rjoerc.in\n"
		"*jpicorc.in\n"
		"-syntax joerc\n"
		"[markdown]\n"
		"*.md\n"
		"-syntax md\n"
		"-wordwrap\n"
		"[yaml]\n"
		"*.yml\n"
		"*.yaml\n"
		"-autoindent\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-pound_comment\n"
		"-syntax yaml\n"
		"[debian]\n"
		"*apt/sources.list\n"
		"*sources.list.*\n"
		"-pound_comment\n"
		"-syntax debian\n"
		"[java_properties]\n"
		"*.properties\n"
		"-encoding iso-8859-1\n"
		"-pound_comment\n"
		"-syntax properties\n"
		"[ini]\n"
		"*.ini\n"
		"*.inf\n"
		"*hgrc\n"
		"*.gitconfig\n"
		"*.git/config\n"
		"-syntax ini\n"
		"-pound_comment\n"
		"-semi_comment\n"
		"[erb]\n"
		"*.erb\n"
		"-syntax erb\n"
		"[erb_html]\n"
		"*.html.erb\n"
		"-syntax htmlerb\n"
		"[haml]\n"
		"*.haml\n"
		"-syntax haml\n"
		"-istep 2\n"
		"-indentc 32\n"
		"-autoindent\n"
		"-smarthome\n"
		"-smartbacks\n"
		"[erlang]\n"
		"*.erl\n"
		"*.eterm\n"
		"*.app.src\n"
		"*rebar.config\n"
		"*.hrl\n"
		"*.erlang\n"
		"-syntax erlang\n"
		"-autoindent\n"
		"-title\n"
		"[sieve]\n"
		"*.sieve\n"
		"-syntax sieve\n"
		"-autoindent\n"
		"-title\n"
		"[puppet]\n"
		"*.pp\n"
		"-syntax puppet\n"
		"-pound_comment\n"
		"-autoindent\n"
		"-title\n"
		"[iptables]\n"
		"*\n"
		"+# Generated by iptables-save\n"
		"+# Generated by ip6tables-save\n"
		"-syntax iptables\n"
		"-pound_comment\n"
		"[powershell]\n"
		"*.ps1\n"
		"*.psm1\n"
		"-syntax powershell\n"
		"-autoindent\n"
		"-smarthome\n"
		"-smartbacks\n"
		"[git-commit]\n"
		"*COMMIT_EDITMSG\n"
		"-syntax git-commit\n"
		"[elixir]\n"
		"*.ex\n"
		"*.exs\n"
		"-autoindent\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-title\n"
		"-syntax elixir\n"
		"-pound_comment\n"
		"-text_delimiters do=else=end\n"
		"-indentc 32\n"
		"-istep 2\n"
		"[r]\n"
		"*.r\n"
		"-autoindent\n"
		"-title\n"
		"-syntax r\n"
		"-pound_comment\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"[groovy]\n"
		"*.gvy\n"
		"*.groovy\n"
		"*.gy\n"
		"*.gsh\n"
		"-autoindent\n"
		"-title\n"
		"-syntax groovy\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"[clojure]\n"
		"*.clj\n"
		"*.cljs\n"
		"*.cljc\n"
		"*.edn\n"
		"-autoindent\n"
		"-title\n"
		"-syntax clojure\n"
		"-semi_comment\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"[rust]\n"
		"*.rs\n"
		"-autoindent\n"
		"-title\n"
		"-syntax rust\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"[coffee]\n"
		"*.coffee\n"
		"-autoindent\n"
		"-title\n"
		"-syntax coffee\n"
		"-pound_comment\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"[scala]\n"
		"*.scala\n"
		"-autoindent\n"
		"-title\n"
		"-syntax scala\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"[swift]\n"
		"*.swift\n"
		"-autoindent\n"
		"-title\n"
		"-syntax swift\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"[d]\n"
		"*.d\n"
		"-autoindent\n"
		"-title\n"
		"-syntax d\n"
		"-c_comment\n"
		"-cpp_comment\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-purify\n"
		"-single_quoted\n"
		"[dockerfile]\n"
		"*Dockerfile\n"
		"-autoindent\n"
		"-title\n"
		"-syntax dockerfile\n"
		"-pound_comment\n"
		"[spec]\n"
		"*.spec\n"
		"-autoindent\n"
		"-syntax spec\n"
		"[haskell]\n"
		"*.hs\n"
		"-autoindent\n"
		"-smarthome\n"
		"-smartbacks\n"
		"-syntax haskell\n"
		"-vhdl_comment\n"
,	"c.jsf",
		"\n"
		"\n"
		"\n"
		"\n"
		"\n"
		"\n"
		"\n"
		"\n"
		"-\n"
		"\n"
		"=Idle\n"
		"=Ident\n"
		"=Bad\n"
		"\n"
		"=Preproc\n"
		"=Precond	+Preproc\n"
		"=Define		+DefinedIdent +Preproc\n"
		"=IncLocal	+String +Preproc\n"
		"=IncSystem	+Preproc +Keyword\n"
		"\n"
		"=Comment\n"
		"=Constant\n"
		"=Number		+Constant\n"
		"=String		+Constant\n"
		"=StringEscape	+Escape +String\n"
		"=Character	+String\n"
		"=CharacterEscape +StringEscape +Character\n"
		"=Escape\n"
		"\n"
		"=Keyword\n"
		"=CppKeyword	+Keyword\n"
		"=Statement	+Keyword\n"
		"=Loop		+Statement\n"
		"=Conditional	+Statement\n"
		"=Label		+DefinedIdent\n"
		"=Type\n"
		"=StorageClass	+Type +Keyword\n"
		"=Structure	+Type +Keyword\n"
		"=Brace\n"
		"=Control\n"
		"\n"
		":reset Idle\n"
		"	*		first		noeat\n"
		"	\" \\t\"		reset\n"
		"\n"
		":first Idle\n"
		"	*		idle		noeat\n"
		"	\"#\"		pre		mark\n"
		"\n"
		":pre Preproc\n"
		"	*		preproc		noeat\n"
		"	\" \\t\"		pre\n"
		"	\"a-z\"		preident	recolor=-1 buffer\n"
		"\n"
		":preident Preproc\n"
		"	*		preproc		noeat markend recolormark strings\n"
		"	\"define\"	predef		markend recolormark\n"
		"	\"include\"	preinc		markend recolormark\n"
		"	\"if\"		precond		markend recolormark\n"
		"	\"ifdef\"		precond		markend recolormark\n"
		"	\"ifndef\"	precond		markend recolormark\n"
		"	\"else\"		precond		markend recolormark\n"
		"	\"elif\"		precond		markend recolormark\n"
		"	\"endif\"		precond		markend recolormark\n"
		"done\n"
		"	\"a-z\"		preident\n"
		"\n"
		":precond Precond\n"
		"	*		preproc		noeat\n"
		"\n"
		":preinc Preproc\n"
		"	*		preinc\n"
		"	\" \\t\"		preinc_ws\n"
		"	\"\\n\"		reset\n"
		"\n"
		":preinc_ws Preproc\n"
		"	*		prebad		recolor=-1\n"
		"	\" \\t\"		preinc_ws\n"
		"	\"\\\"\"		preinc_local	recolor=-1\n"
		"	\"<\"		preinc_system	recolor=-1\n"
		"\n"
		":preinc_local IncLocal string\n"
		"	*		preinc_local\n"
		"	\"\\\"\\n\"		reset\n"
		"	\n"
		":preinc_system IncSystem string\n"
		"	*		preinc_system\n"
		"	\">\\n\"		reset\n"
		"\n"
		":prebad	Bad	\n"
		"	*		prebad\n"
		"	\"\\n\"		reset\n"
		"\n"
		"\n"
		":predef Preproc\n"
		"	*		predef\n"
		"	\" \\t\"		predef_ws\n"
		"	\"\\n\"		reset\n"
		"	\n"
		":predef_ws Preproc\n"
		"	*		prebad		recolor=-1\n"
		"	\" \\t\"		predef_ws\n"
		"	\"\\c\"		predef_ident	recolor=-1\n"
		"\n"
		":predef_ident Define\n"
		"	*		idle		noeat\n"
		"	\"\\c\"		predef_ident\n"
		"\n"
		":preproc Preproc\n"
		"	*		preproc\n"
		"	\"\\n\"		reset\n"
		"	\"\\\\\"		preproc_cont\n"
		"	\"/\"		preproc_slash\n"
		"	\n"
		":preproc_slash Preproc\n"
		"	*		preproc		noeat\n"
		"	\"*\"		comment		recolor=-2\n"
		"	\"/\"		line_comment	recolor=-2\n"
		"	\n"
		":preproc_cont Preproc\n"
		"	*		preproc_cont\n"
		"	\"\\n\"		preproc\n"
		"\n"
		"\n"
		":idle Idle\n"
		"	*		idle\n"
		"	\"\\n\"		reset\n"
		"	\"/\"		slash\n"
		"	\"0\"		first_digit	recolor=-1\n"
		"	\"1-9\"		decimal		recolor=-1\n"
		"	\".\"		maybe_float\n"
		"	\"\\\"\"		string		recolor=-1\n"
		"	\"'\"		char		recolor=-1\n"
		"	\"\\i\"		ident		recolor=-1 buffer\n"
		"	\"\\\\\"		outside_escape	recolor=-1\n"
		"	\"{}\"		brace		recolor=-1\n"
		"	\",:;=()><[]*&|!~+\\-%^\"	control	recolor=-1\n"
		"\n"
		":outside_escape	Escape\n"
		"	*	idle\n"
		"\n"
		":brace Brace\n"
		"	*	idle	noeat\n"
		"\n"
		":control Control\n"
		"	*	idle	noeat\n"
		"\n"
		":slash Idle\n"
		"	*		idle		noeat recolor=-2	# Not sure about this\n"
		"	\"*\"		comment		recolor=-2\n"
		"	\"/\"		line_comment	recolor=-2\n"
		"\n"
		":comment Comment comment\n"
		"	*		comment\n"
		"	\"BFHNTX\"	comment		noeat call=comment_todo.comment_todo()\n"
		"	\"*\"		maybe_end_comment\n"
		"\n"
		":maybe_end_comment Comment comment\n"
		"	*		comment\n"
		"	\"/\"		idle\n"
		"	\"*\"		maybe_end_comment\n"
		"\n"
		":line_comment Comment comment\n"
		"	*		line_comment\n"
		"	\"BFHNTX\"	line_comment	noeat call=comment_todo.comment_todo()\n"
		"	\"\\n\"		reset\n"
		"\n"
		":first_digit Number\n"
		"	*		idle	noeat\n"
		"	\"xX\"		hex		recolor=-2\n"
		"	\".\"		float\n"
		"	\"eE\"		epart\n"
		"	\"0-7\"		octal		recolor=-2\n"
		"	\"89\"		bad_number	recolor=-1\n"
		"\n"
		":bad_number Bad\n"
		"	*		idle	noeat\n"
		"	\"0-9\"		bad_number\n"
		"\n"
		":octal Number\n"
		"	*		idle	noeat\n"
		"	\"0-7\"		octal\n"
		"	\"89\"		bad_number	recolor=-1\n"
		"\n"
		":hex Number\n"
		"	*		idle	noeat\n"
		"	\"0-9A-Fa-f\"	hex\n"
		"\n"
		":decimal Number\n"
		"	*		idle	noeat\n"
		"	\"0-9\"		decimal\n"
		"	\"eE\"		epart\n"
		"	\".\"		float\n"
		"\n"
		":maybe_float Number\n"
		"	*		idle	recolor=-2	noeat\n"
		"	\"0-9\"		float		recolor=-2\n"
		"\n"
		":float Number\n"
		"	*		idle	noeat\n"
		"	\"eE\"		epart\n"
		"	\"0-9\"		float\n"
		"\n"
		":epart Number\n"
		"	*		idle	noeat\n"
		"	\"0-9+\\-\"	enum\n"
		"\n"
		":enum Number\n"
		"	*		idle	noeat\n"
		"	\"0-9\"		enum\n"
		"\n"
		":string	String string\n"
		"	*		string\n"
		"	\"\\\"\"		idle\n"
		"	\"\\\\\"		string_escape	recolor=-1\n"
		"	\"%\"		string_control	recolor=-1\n"
		"\n"
		":string_escape StringEscape string\n"
		"	*		string\n"
		"	\"x\"		string_hex\n"
		"	\"0-7\"		string_octal2\n"
		"	\"\\n\"		string		recolor=-2\n"
		"\n"
		":string_hex StringEscape string\n"
		"	*		string		noeat\n"
		"	\"0-9a-fA-F\"	string_hex\n"
		"\n"
		":string_octal2 StringEscape string\n"
		"	*		string		noeat\n"
		"	\"0-7\"		string_octal3\n"
		"\n"
		":string_octal3 StringEscape string\n"
		"	*		string		noeat\n"
		"	\"0-7\"		string\n"
		"\n"
		":string_control StringEscape string\n"
		"	*		string\n"
		"	\"\\\"\"		string noeat\n"
		"	\"\\n\"		reset\n"
		"	\"\\\\\"		string_escape	recolor=-1\n"
		"	\"0-9.\\-+ #hjILtz$\"	string_control\n"
		"\n"
		":char Character string\n"
		"	*		char\n"
		"	\"\\n\"		reset\n"
		"	\"'\"		idle\n"
		"	\"\\\\\"		char_escape	recolor=-1\n"
		"\n"
		":char_escape CharacterEscape string\n"
		"	*		char\n"
		"	\"x\"		char_hex\n"
		"	\"0-7\"		char_octal2\n"
		"	\"\\n\"		char		recolor=-2\n"
		"\n"
		":char_hex CharacterEscape string\n"
		"	*		char		noeat\n"
		"	\"0-9a-fA-F\"	char_hex\n"
		"\n"
		":char_octal2 CharacterEscape string\n"
		"	*		char		noeat\n"
		"	\"0-7\"		char_octal3\n"
		"\n"
		":char_octal3 CharacterEscape string\n"
		"	*		char		noeat\n"
		"	\"0-7\"		char\n"
		"\n"
		":ident Ident\n"
		"	*		idle		noeat strings\n"
		"	\"int\"		type\n"
		"	\"float\"		type\n"
		"	\"long\"		type\n"
		"	\"short\"		type\n"
		"	\"char\"		type\n"
		"	\"double\"	type\n"
		"	\"signed\"	type\n"
		"	\"unsigned\"	type\n"
		"	\"void\"		type\n"
		"	\"static\"	storage\n"
		"	\"register\"	storage\n"
		"	\"extern\"	storage\n"
		"	\"inline\"	type\n"
		"	\"auto\"		storage\n"
		"	\"const\"		storage\n"
		"	\"volatile\"	storage\n"
		"	\"if\"		cond\n"
		"	\"else\"		cond\n"
		"	\"while\"		loop\n"
		"	\"for\"		loop\n"
		"	\"break\"		stmt\n"
		"	\"continue\"	stmt\n"
		"	\"do\"		loop\n"
		"	\"case\"		label\n"
		"	\"default\"	label\n"
		"	\"switch\"	cond\n"
		"	\"goto\"		stmt\n"
		"	\"struct\"	struct\n"
		"	\"enum\"		struct\n"
		"	\"return\"	stmt\n"
		"	\"sizeof\"	kw\n"
		"	\"typedef\"	struct\n"
		"	\"union\"		struct\n"
		"	\"asm\"		kw\n"
		"	#\"asm\"			cppkw (listed above as a C keyword)\n"
		"	\"bool\"			cppkw\n"
		"	\"catch\"			cppkw\n"
		"	\"class\"			struct\n"
		"	\"const_cast\"		cppkw\n"
		"	\"delete\"		cppkw\n"
		"	\"dynamic_cast\"		cppkw\n"
		"	\"explicit\"		cppkw\n"
		"	\"false\"			cppkw\n"
		"	\"friend\"		cppkw\n"
		"	#\"inline\"		cppkw (listed above as a C keyword)\n"
		"	\"mutable\"		cppkw\n"
		"	\"namespace\"		cppkw\n"
		"	\"new\"			cppkw\n"
		"	\"operator\"		cppkw\n"
		"	\"private\"		cppkw\n"
		"	\"protected\"		cppkw\n"
		"	\"public\"		cppkw\n"
		"	\"reinterpret_cast\"	cppkw\n"
		"	\"static_cast\"		cppkw\n"
		"	\"template\"		cppkw\n"
		"	\"this\"			cppkw\n"
		"	\"throw\"			cppkw\n"
		"	\"true\"			cppkw\n"
		"	\"try\"			cppkw\n"
		"	\"typeid\"		cppkw\n"
		"	\"typename\"		cppkw\n"
		"	\"using\"			cppkw\n"
		"	\"virtual\"		cppkw\n"
		"	\"wchar_t\"		type\n"
		"	\"typeof\"		cppkw\n"
		"done\n"
		"	\"\\c\"		ident\n"
		"\n"
		":type Type\n"
		"	*	idle	noeat\n"
		"\n"
		":kw Keyword\n"
		"	*	idle	noeat\n"
		"\n"
		":cppkw CppKeyword\n"
		"	*	idle	noeat\n"
		"\n"
		":label Label\n"
		"	*	idle	noeat\n"
		"\n"
		":loop Loop\n"
		"	*	idle	noeat\n"
		"\n"
		":stmt Statement\n"
		"	*	idle	noeat\n"
		"\n"
		":storage StorageClass\n"
		"	*	idle	noeat\n"
		"\n"
		":cond Conditional\n"
		"	*	idle	noeat\n"
		"\n"
		":struct Structure\n"
		"	*	idle	noeat\n"
,	"default.jcf",
		".colors 16\n"
		"-selection 	default/default inverse\n"
		"-menusel 	default/default inverse\n"
		"-linum		default dim\n"
		"-curlinum	default bold\n"
		"=Idle\n"
		"=Bad            bold red\n"
		"=Preproc        blue\n"
		"=Define         bold blue\n"
		"=Comment        green\n"
		"=IncLocal       cyan\n"
		"=IncSystem      bold cyan\n"
		"=Constant       cyan\n"
		"=Escape         bold cyan\n"
		"=Type           bold\n"
		"=Keyword        bold\n"
		"=Brace          magenta\n"
		"=Control\n"
		"=Ident\n"
		"=Tag		bold blue\n"
		"=TagName	+Tag\n"
		"=TODO		default/default inverse\n"
		"=Title		bold\n"
		"=diff.DelLine	red\n"
		"=diff.AddLine	green\n"
		"=diff.ChgLine	magenta\n"
		"=diff.Garbage	dim white\n"
		"=diff.FileOld	bold red\n"
		"=diff.FileNew	bold green\n"
		"=diff.Hunk	bold blue\n"
		"# For backwards compatibility with older syntaxes that specify colors,\n"
		"# the scheme must specify termcolors.  Yes this is a bit of a kludge.\n"
		"-term  0 0\n"
		"-term  1 1\n"
		"-term  2 2\n"
		"-term  3 3\n"
		"-term  4 4\n"
		"-term  5 5\n"
		"-term  6 6\n"
		"-term  7 7\n"
		"-term  8 8\n"
		"-term  9 9\n"
		"-term 10 10\n"
		"-term 11 11\n"
		"-term 12 12\n"
		"-term 13 13\n"
		"-term 14 14\n"
		"-term 15 15\n"
,	"comment_todo.jsf",
		"\n"
		"=Comment 	\n"
		"=TODO		+Comment\n"
		"\n"
		".subr comment_todo\n"
		":comment_todo_init Comment comment\n"
		"	*		comment_todo_guess	buffer\n"
		"\n"
		":comment_todo_guess Comment comment\n"
		"	*		comment_todo_unknown	noeat strings\n"
		"	\"BUG\"		comment_todo\n"
		"	\"FIXME\"		comment_todo\n"
		"	\"HACK\"		comment_todo\n"
		"	\"NOTE\"		comment_todo\n"
		"	\"TODO\"		comment_todo\n"
		"	\"XXX\"		comment_todo\n"
		"done\n"
		"	\"A-Z\"		comment_todo_guess\n"
		"\n"
		":comment_todo_unknown Comment comment\n"
		"	*		NULL		noeat return\n"
		"\n"
		":comment_todo TODO comment\n"
		"	*		NULL		noeat return\n"
		".end\n"
,	"filename.jsf",
		"\n"
		"=Idle\n"
		"=Notused	+Comment\n"
		"\n"
		":idle Idle\n"
		"	*	filename	noeat mark\n"
		"\n"
		":filename Idle\n"
		"	*	filename\n"
		"	\"/\"	slash\n"
		"	\"\\n\"	idle\n"
		"\n"
		":slash Idle\n"
		"	*	filename\n"
		"	\"/\"	restart		noeat recolormark\n"
		"	\"~\"	restart		noeat recolormark\n"
		"	\"\\n\"	idle\n"
		"\n"
		":restart Notused\n"
		"	*	filename	noeat mark\n"
,	"context.jsf",
		"\n"
		"=Idle\n"
		"=Title bold # Doesn't matter, just make it visible\n"
		"\n"
		":idle Idle\n"
		"	*		title_line	buffer noeat recolor=-1\n"
		"	\"b\"		maybe_begin_1	buffer\n"
		"	\"B\"		maybe_BEGIN_1	buffer\n"
		"	\"\\n\"		idle\n"
		"	\"/\"		maybe_comment	buffer\n"
		"	\"-\"		maybe_comment_1	buffer\n"
		"	\";{} \\t\\r\\f#\"	comment\n"
		"\n"
		":maybe_comment Idle\n"
		"	*		title_line	noeat recolor=-1\n"
		"	\"/*\"		comment\n"
		"\n"
		":maybe_comment_1 Idle\n"
		"	*		title_line	noeat recolor=-1\n"
		"	\"-\"		comment\n"
		"\n"
		":title_line Title\n"
		"	*		title_line\n"
		"	\"\\n\"		idle		save_s\n"
		"\n"
		":comment Idle\n"
		"	*		comment\n"
		"	\"\\n\"		idle\n"
		"\n"
		":maybe_begin_1 Idle\n"
		"	*		title_line	noeat recolor=-2\n"
		"	\"e\"		maybe_begin_2\n"
		"\n"
		":maybe_begin_2 Idle\n"
		"	*		title_line	noeat recolor=-3\n"
		"	\"g\"		maybe_begin_3\n"
		"\n"
		":maybe_begin_3 Idle\n"
		"	*		title_line	noeat recolor=-4\n"
		"	\"i\"		maybe_begin_4\n"
		"\n"
		":maybe_begin_4 Idle\n"
		"	*		title_line	noeat recolor=-5\n"
		"	\"n\"		comment\n"
		"\n"
		":maybe_BEGIN_1 Idle\n"
		"	*		title_line	noeat recolor=-2\n"
		"	\"E\"		maybe_BEGIN_2\n"
		"\n"
		":maybe_BEGIN_2 Idle\n"
		"	*		title_line	noeat recolor=-3\n"
		"	\"G\"		maybe_BEGIN_3\n"
		"\n"
		":maybe_BEGIN_3 Idle\n"
		"	*		title_line	noeat recolor=-4\n"
		"	\"I\"		maybe_BEGIN_4\n"
		"\n"
		":maybe_BEGIN_4 Idle\n"
		"	*		title_line	noeat recolor=-5\n"
		"	\"N\"		comment\n"
,	NULL
};
