/*
 *	JOE state file
 *	Copyright
 *		(C) 1992 Joseph H. Allen
 *
 *	This file is part of JOE (Joe's Own Editor)
 */

/* Set to enable use of ~/.joe_state file */

extern int joe_state;

/* Load state from file */

void load_state(void);

/* Save state to file */

void save_state(void);
