/*
 *	Position history
 *	Copyright
 *		(C) 1992 Joseph H. Allen
 *
 *	This file is part of JOE (Joe's Own Editor)
 */

void afterpos(void);
void aftermove(W *w, P *p);
void windie(W *w);
int uprevpos(W *w, int k);
int unextpos(W *w, int k);
