/*
 *	File selection menu
 *	Copyright
 *		(C) 1992 Joseph H. Allen
 *
 *	This file is part of JOE (Joe's Own Editor)
 */

int cmplt_file(BW *bw, int k);
int cmplt_file_in(BW *bw, int k);
int cmplt_file_out(BW *bw, int k);
int cmplt_command(BW *bw, int k);
extern int menu_jump;
extern int menu_explorer;
