/*
 *	tags file symbol lookup
 *	Copyright
 *		(C) 1992 Joseph H. Allen
 *
 *	This file is part of JOE (Joe's Own Editor)
 */

extern int notagsmenu;
int utagjump(W *w, int k);
int utag(W *w, int k);
